require 'cell/meta'
require 'active_support/callbacks'

module Cell
  module Context
    THREAD_KEY = :'Cell.cell_id'

    module ClassMethods
      def self.extended(cls)
        cls.include ::ActiveSupport::Callbacks
        cls.define_callbacks :set_tenant, :use_tenant
      end

      def after_set(*args, &block)
        set_callback(:set_tenant, :after, *args, &block)
      end

      def around_use(*args, &block)
        set_callback(:use_tenant, :around, *args, &block)
      end

      def current
        Thread.current[THREAD_KEY]
      end

      def set!(tenant, exclusive: false)
        Meta.with_schema(tenant.schema_name, exclusive: exclusive)
        set_current(tenant)
        tenant.run_callbacks :set_tenant
        tenant
      end

      def use(tenant, exclusive: false)
        fail ArgumentError, "block required" unless block_given?
        saved_tenant = current

        if tenant
          Meta.with_schema(tenant.schema_name, exclusive: exclusive) do
            set_current(tenant)
            tenant.run_callbacks :use_tenant do
              yield Model.current
            end
          end
        else
          Meta.with_global_schema do
            set_current(nil)
            yield nil
          end
        end

        tenant
      ensure
        set_current(saved_tenant)
      end

      private
      def set_current(tenant)
        unless tenant.nil? || tenant.is_a?(Model)
          fail ArgumentError, "Invalid tenant: #{tenant.inspect}"
        end

        Thread.current[THREAD_KEY] = tenant
      end
    end

    def use(*args, &block)
      self.class.use(self, *args, &block)
    end

    def set!(exclusive: false)
      self.class.set!(self, exclusive: exclusive)
    end

    def current?
      self.class.current == self
    end

    def self.prepended(cls)
      cls.extend(ClassMethods)
    end
  end
end
