# These are added to ActiveRecord::Base, regardless if they're the tenant model or not.

# There are two big tricks here:
#  1. Assigning a "tenant" attribute to every model that is loaded from the db, or created while a
#     tenant is active.
#  2. Making sure the object is validated and saved with the proper tenant activated, even if the
#     tenant has changed since the object was created
#
# So you can load an object from "tenant_a", switch to "tenant_b", and if you save or validate the
# object, it'll do so while switched to "tenant_a", and then restore.

require 'cell/meta'

module Cell
  module Ext
    module ActiveRecord
      # Say we have a model 'u', that is from an old Tenant, loading an association like 'u.books'
      # should swap into u.tenant while fetching.
      module Association
        def load_target
          owner._activate_tenant do
            super
          end
        end
      end

      module ClassMethods
        def global_model?
          Meta.global_model?(self)
        end

        # When a model's schema is lazy-loaded, we want to make sure our view of the DB allows us
        # to get introspection data.
        def load_schema!
          Meta.with_structural_schema do
            return super
          end
        end
      end

      attr_reader :tenant

      def _assign_tenant
        unless self.class.global_model?
          @tenant = Model.current
        end
      end

      def _activate_tenant(&block)
        if tenant && tenant != Model.current
          tenant.use(&block)
        else
          yield
        end
      end

      # undocumented, but makes sure validations are ran in the context of the object's tenant,
      # e.g., uniqueness.
      def run_validations!
        _activate_tenant do
          super
        end
      end

      def self.prepended(cls)
        class << cls
          prepend ClassMethods
        end

        cls.after_initialize :_assign_tenant
        cls.around_save :_activate_tenant
      end
    end
  end
end

ActiveSupport.on_load(:active_record) do
  ActiveRecord::Base.prepend(Cell::Ext::ActiveRecord)
  ActiveRecord::Associations::Association.prepend(Cell::Ext::ActiveRecord::Association)
  ActiveRecord::Associations::CollectionAssociation.prepend(Cell::Ext::ActiveRecord::Association)
end
